{ Copyright (C) 1998-2023, written by Shkolnik Mike, Scalabium
  E-Mail:  mshkolnik@scalabium.com
           mshkolnik@yahoo.com
  WEB: http://www.scalabium.com

  TSMButtonBar component is the panel with list of buttons
}
unit SMButtonBar         ;

interface

{$I SMVersion.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Buttons,
  ExtCtrls;

type
  TSMButtonInBar = class(TCollectionItem)
  private
    FImageIndex: Integer;

    function GetCaption: TCaption;
    procedure SetCaption(AValue: TCaption);

    function GetEnabled: Boolean;
    procedure SetEnabled(AValue: Boolean);

    function GetFont: TFont;
    procedure SetFont(AValue: TFont);

    function GetGlyph: TBitmap;
    procedure SetGlyph(AValue: TBitmap);

    function GetHint: string;
    procedure SetHint(AValue: string);

    procedure SetImageIndex(AValue: Integer);

    function GetLayout: TButtonLayout;
    procedure SetLayout(AValue: TButtonLayout);

    function GetParentFont: Boolean;
    procedure SetParentFont(AValue: Boolean);

    function GetShowHint: Boolean;
    procedure SetShowHint(AValue: Boolean);

    function GetVisible: Boolean;
    procedure SetVisible(AValue: Boolean);

    function GetOnClick: TNotifyEvent;
    procedure SetOnClick(AValue: TNotifyEvent);
  protected
    FButton: TBitBtn;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
  published
    property Caption: TCaption read GetCaption write SetCaption;
    property Enabled: Boolean read GetEnabled write SetEnabled;
    property Font: TFont read GetFont write SetFont;
    property Glyph: TBitmap read GetGlyph write SetGlyph;
    property Hint: string read GetHint write SetHint;
    property ImageIndex: Integer read FImageIndex write SetImageIndex;
    property Layout: TButtonLayout read GetLayout write SetLayout;
    property ParentFont: Boolean read GetParentFont write SetParentFont;
    property ShowHint: Boolean read GetShowHint write SetShowHint;
    property Visible: Boolean read GetVisible write SetVisible;

    property OnClick: TNotifyEvent read GetOnClick write SetOnClick;
  end;

  TSMBarButtons = class(TCollection)
  private
    FSMBarControl: TComponent;

    function GetButton(Index: Integer): TSMButtonInBar;
    procedure SetButton(Index: Integer; Value: TSMButtonInBar);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(SMBarCntrl: TComponent);
    function Add: TSMButtonInBar;

    property Items[Index: Integer]: TSMButtonInBar read GetButton write SetButton; default;
    property ButtonBarControl: TComponent read FSMBarControl;
  end;

  {$IFDEF SM_ADD_ComponentPlatformsAttribute}
  [ComponentPlatformsAttribute(pidWin32 or pidWin64)]
  {$ENDIF}
  TSMButtonBar = class(TPanel)
  private
    { Private declarations }
    FButtons: TSMBarButtons;

    FButtonHeight: Integer;
    FButtonWidth: Integer;

    FImageList: TImageList;

    function GetImageList: TImageList;
    procedure SetImageList(AValue: TImageList);

    procedure SetButtons(AValue: TSMBarButtons);
    procedure SetButtonHeight(AValue: Integer);
    procedure SetButtonWidth(AValue: Integer);
  protected
    { Protected declarations }
    procedure CreateDefaultButtons;

    procedure ApplyValuesToButtons;
    function GetLastRightPosition(ALastIndex: Integer): Integer;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    { Published declarations }
    property Buttons: TSMBarButtons read FButtons write SetButtons;
    property ButtonHeight: Integer read FButtonHeight write SetButtonHeight;
    property ButtonWidth: Integer read FButtonWidth write SetButtonWidth;

    property ImageList: TImageList read GetImageList write SetImageList;
  end;

implementation
{$R *.RES}
{$R smbb.RES}

const
  ButtonMargin = 2;

{ TSMButtonInBar }
constructor TSMButtonInBar.Create(Collection: TCollection);
var
  intLeft: Integer;
begin
  inherited Create(Collection);

  FImageIndex := -1;

  intLeft := (TSMBarButtons(Collection).ButtonBarControl as TSMButtonBar).GetLastRightPosition(-1);
  FButton := TBitBtn.Create(TSMBarButtons(Collection).ButtonBarControl);
  FButton.Parent := TSMBarButtons(Collection).ButtonBarControl as TSMButtonBar;
  FButton.Left := intLeft;
  FButton.Top := ButtonMargin;
end;

destructor TSMButtonInBar.Destroy;
begin
  if Assigned(FButton) then
  begin
    FButton.Free;
    FButton := nil;
  end;

  inherited Destroy;
end;

procedure TSMButtonInBar.Assign(Source: TPersistent);
var
  b: TSMButtonInBar;
begin
  if Source is TSMButtonInBar then
  begin
    if Assigned(Collection) then
      Collection.BeginUpdate;
    try
      b := TSMButtonInBar(Source);
      FButton.Assign(b.FButton)
    finally
      if Assigned(Collection) then
        Collection.EndUpdate;
    end;
  end
  else
    inherited Assign(Source);
end;

function TSMButtonInBar.GetCaption: TCaption;
begin
  if Assigned(FButton) then
    Result := FButton.Caption
end;

procedure TSMButtonInBar.SetCaption(AValue: TCaption);
begin
  if Assigned(FButton) then
    FButton.Caption := AValue
end;

function TSMButtonInBar.GetEnabled: Boolean;
begin
  if Assigned(FButton) then
    Result := FButton.Enabled
  else
    Result := False;
end;

procedure TSMButtonInBar.SetEnabled(AValue: Boolean);
begin
  if Assigned(FButton) then
    FButton.Enabled := AValue
end;

function TSMButtonInBar.GetFont: TFont;
begin
  if Assigned(FButton) then
    Result := FButton.Font
  else
    Result := nil
end;

procedure TSMButtonInBar.SetFont(AValue: TFont);
begin
  if Assigned(FButton) then
    FButton.Font.Assign(AValue)
end;

function TSMButtonInBar.GetGlyph: TBitmap;
begin
  if Assigned(FButton) then
    Result := FButton.Glyph
  else
    Result := nil
end;

procedure TSMButtonInBar.SetGlyph(AValue: TBitmap);
begin
  if Assigned(FButton) then
    FButton.Glyph.Assign(AValue)
end;

function TSMButtonInBar.GetHint: string;
begin
  if Assigned(FButton) then
    Result := FButton.Hint
  else
    Result := ''
end;

procedure TSMButtonInBar.SetHint(AValue: string);
begin
  if Assigned(FButton) then
    FButton.Hint := AValue
end;

procedure TSMButtonInBar.SetImageIndex(AValue: Integer);
begin
  if Assigned(FButton) then
    FImageIndex := AValue
end;

function TSMButtonInBar.GetLayout: TButtonLayout;
begin
  if Assigned(FButton) then
    Result := FButton.Layout
  else
    Result := blGlyphLeft
end;

procedure TSMButtonInBar.SetLayout(AValue: TButtonLayout);
begin
  if Assigned(FButton) then
    FButton.Layout := AValue
end;

function TSMButtonInBar.GetParentFont: Boolean;
begin
  if Assigned(FButton) then
    Result := FButton.ParentFont
  else
    Result := False
end;

procedure TSMButtonInBar.SetParentFont(AValue: Boolean);
begin
  if Assigned(FButton) then
    FButton.ParentFont := AValue
end;

function TSMButtonInBar.GetShowHint: Boolean;
begin
  if Assigned(FButton) then
    Result := FButton.ShowHint
  else
    Result := False
end;

procedure TSMButtonInBar.SetShowHint(AValue: Boolean);
begin
  if Assigned(FButton) then
    FButton.ShowHint := AValue
end;

function TSMButtonInBar.GetVisible: Boolean;
begin
  if Assigned(FButton) then
    Result := FButton.Visible
  else
    Result := False
end;

procedure TSMButtonInBar.SetVisible(AValue: Boolean);
begin
  if Assigned(FButton) then
    FButton.Visible := AValue
end;

function TSMButtonInBar.GetOnClick: TNotifyEvent;
begin
  if Assigned(FButton) then
    Result := FButton.OnClick
  else
    Result := nil
end;

procedure TSMButtonInBar.SetOnClick(AValue: TNotifyEvent);
begin
  if Assigned(FButton) then
    FButton.OnClick := AValue
end;

{ TSMBarButtons }
constructor TSMBarButtons.Create(SMBarCntrl: TComponent);
begin
  inherited Create(TSMButtonInBar);

  FSMBarControl := SMBarCntrl;
end;

function TSMBarButtons.Add: TSMButtonInBar;
begin
  Result := TSMButtonInBar(inherited Add);
end;

function TSMBarButtons.GetButton(Index: Integer): TSMButtonInBar;
begin
  Result := TSMButtonInBar(inherited Items[Index]);
end;

procedure TSMBarButtons.SetButton(Index: Integer; Value: TSMButtonInBar);
begin
  Items[Index].Assign(Value);
end;

function TSMBarButtons.GetOwner: TPersistent;
begin
  Result := FSMBarControl;
end;

{ TSMButtonBar }
constructor TSMButtonBar.Create(AOwner: TComponent);
begin
  inherited;

  FButtons := TSMBarButtons.Create(Self);

  FButtonHeight := 30;
  FButtonWidth := 30;
  Caption := ' ';

  CreateDefaultButtons;
  ApplyValuesToButtons;
end;

destructor TSMButtonBar.Destroy;
begin
  if Assigned(FButtons) then
  begin
    FButtons.Free;
    FButtons := nil;
  end;

  inherited;
end;

procedure TSMButtonBar.CreateDefaultButtons;

  procedure LoadImage(const AName: string; bbtn: TSMButtonInBar);
  begin
    bbtn.Glyph.LoadFromResourceName(HInstance, 'smbb_' + AName);
  end;

const
  arrNames: array[0..15] of string = ('First', 'Prev', 'Next', 'Last', 'New', 'Copy', 'Delete', 'Edit', 'Filter', 'Find',
                                      'Print', 'Export', 'Import', 'Post', 'Cancel', 'Refresh');
var
  bbtn: TSMButtonInBar;
  i: Integer;
begin
  for i := 0 to 15 do
  begin
    bbtn := Buttons.Add;
//    bbtn.Caption := arrNames[i];
    bbtn.Hint := arrNames[i];
    LoadImage(arrNames[i], bbtn);
  end;
end;

procedure TSMButtonBar.SetButtons(AValue: TSMBarButtons);
begin
  if Assigned(FButtons) then
    FButtons.Assign(AValue)
end;

function TSMButtonBar.GetImageList: TImageList;
begin
  Result := FImageList;
end;

procedure TSMButtonBar.SetImageList(AValue: TImageList);
begin
{$IFDEF SMForDelphi5}
  if FImageList <> nil then
    FImageList.RemoveFreeNotification(Self);
{$ENDIF}

  FImageList := AValue;
  ApplyValuesToButtons;

  if AValue <> nil then
    AValue.FreeNotification(Self);
end;

procedure TSMButtonBar.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);

  if (Operation = opRemove) then
  begin
    if (AComponent = FImageList) then
      ImageList := nil;
  end;
end;

procedure TSMButtonBar.SetButtonHeight(AValue: Integer);
begin
  FButtonHeight := AValue;
  ApplyValuesToButtons;
end;

procedure TSMButtonBar.SetButtonWidth(AValue: Integer);
begin
  FButtonWidth := AValue;
  ApplyValuesToButtons;
end;

procedure TSMButtonBar.ApplyValuesToButtons;
var
  i: Integer;
begin
  if not Assigned(FButtons) then exit;

  for i := 0 to FButtons.Count-1 do
  begin
    FButtons[i].FButton.Height := FButtonHeight;
    FButtons[i].FButton.Width := FButtonWidth;

    if Assigned(ImageList) and (FButtons[i].ImageIndex >= 0)then
      ImageList.GetBitmap(FButtons[i].ImageIndex, FButtons[i].FButton.Glyph);

    FButtons[i].FButton.Left := GetLastRightPosition(i);
  end;
end;

function TSMButtonBar.GetLastRightPosition(ALastIndex: Integer): Integer;
var
  btn: TBitBtn;
begin
  Result := ButtonMargin;
  if not Assigned(FButtons) then exit;

  if (ALastIndex < 0) then
    ALastIndex := FButtons.Count-1;
  if (ALastIndex > 0) then
  begin
    btn := FButtons[ALastIndex-1].FButton;
    if Assigned(btn) then
      Result := btn.Left + btn.Width + ButtonMargin;
  end
  else
    Result := ButtonMargin;
end;

end.
